
/* Network traffic analyzer
 * (c) 2001, Emil Mikulic.
 */

/* Horribly incorrect and minimalistic web server */

#include "www.h"
#include "host_db.h"
#include "port_db.h"
#include "proto_db.h"
#include "graph.h"
#include "gif.h"
#include "content.h"

#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>



#define error(x) freakout("WWW: " x)
#define POLLDELAY 5 /* seconds */
#define MAXLISTEN 10

pthread_t www_thread;



inline void html_graph(content *cnt, char *title, int GRAPHWIDTH, int GRAPHHEIGHT,
		i64 graph_in[], i64 graph_out[], int pos, int GRAPH_SIZE)
{
	int i, total_width = 0, total_strips = 0;
	double divisor;
	i64 max_transfer = {0, 0};
	char *maximum;
	
	/* find maximum value for this graph */
	for (i=0; i<GRAPH_SIZE; i++)
		max_transfer = max_64(max_transfer,
				max_64(graph_in[i],graph_out[i]));

	maximum = add_commas( bignum_from_64bit(max_transfer), 1 );

	c_append(cnt,
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">"
		"<tr><td bgcolor=\"#000000\">"
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">"
		"<tr><td align=\"center\">"
		"<font color=\"#FFFFFF\"><b>%s</b> "
		"(<font color=\"#FF0000\">in</font> <font color=\"#FF00FF\">and</font> "
		"<font color=\"#0000FF\">out</font>)<br>"
		"Maximum: %s bytes</td></tr>"
		"<tr><td bgcolor=\"#FFFFFF\">"
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\"><tr>",
		title, maximum);

	free(maximum);

	divisor = max(1.0, db_div_64_d(max_transfer, GRAPHHEIGHT));

	i = pos;
	do {
		int in, out, blank, shared, taint = 0, width;

		i++;
		if (i == GRAPH_SIZE) i = 0;

		/* width calculation */
		width = (GRAPHWIDTH - total_width - 2*GRAPH_SIZE) / (GRAPH_SIZE - total_strips);
		total_strips++;
		total_width += width;

		/* get data */
		in = i_div_64_db(graph_in[i], divisor);
		out = i_div_64_db(graph_out[i], divisor);
		blank = GRAPHHEIGHT - max(in, out);
		shared = min(in, out);

		/* write it out */
		c_append(cnt, "<td valign=\"bottom\">");
		if (blank)
		{
			taint = 1;
			c_append(cnt, "<img src=\"i/x\" width=\"%d\" height=\"%d\">",
					width, blank);
		}
		if ((in > out) && (in-shared))
		{
			c_append(cnt, "%s<img src=\"i/r\" width=\"%d\" height=\"%d\">",
					taint?"<br>":"", width, in-shared);
			taint = 1;
		}
		else if (out-shared)
		{
			c_append(cnt, "%s<img src=\"i/b\" width=\"%d\" height=\"%d\">",
					taint?"<br>":"", width, out-shared);
			taint = 1;
		}
		if (shared)
			c_append(cnt, "%s<img src=\"i/m\" width=\"%d\" height=\"%d\">",
					taint?"<br>":"", width, shared);

		c_append(cnt, "</td>");

	} while (i != pos);

	c_append(cnt,
		"</tr></table>"
		"</td></tr></table></td></tr></table>");
}



const char css[] = "<style type=\"text/css\">"
	"A{color:#404040;text-decoration:none;}"
	"A:hover{color:#000000;background:#E0E0E0; text-decoration:underline;}"
	"A.black{color:#FFFFFF;text-decoration:underline;}"
	"A.black:hover{color:#FFFFFF;background:#808080; text-decoration:underline;}"
	"</style>";

/* J.P. used pipes ('|') instead of greater-thans ('>' = &gt;)
 * in a screenshot he sent me.  I thought the pipes looked good.
 */
const char linkbox[] = "<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\">"
	"<tr><td bgcolor=\"#000000\"><font color=\"#FFFFFF\"><b>"
	"&nbsp;" PACKAGE " v" VERSION "&nbsp;"
	"</b></font></td>"
	"<td bgcolor=\"#000000\">"
	"<table border=\"0\" cellspacing=\"1\" cellpadding=\"2\">"
	"<tr><td bgcolor=\"#FFFFFF\">"
	"<a href=\"/\">main</a> | "
	"<a href=\"hosts-total.html\">hosts</a> | "
	"<a href=\"ports-total.html\">ports</a> | "
	"<a href=\"protocols.html\">protocols</a> | "
	"<a href=\"graphs.html\">graphs</a> | "
	"<a href=\"" URL "\">homepage</a>"
	"</td></tr></table>"
	"</td></tr></table>";

content *www_frontpage(SOCKET incoming)
{
	char *x_packets, *x_data;
	extern i64 num_packets, total_data;
	extern time_t t_start, t_already;
	content *cnt = c_new();

	time_t uptime = t_already + (time(NULL) - t_start);
	dword days, hours, minutes;
	
	c_append(cnt, "<html><head><title>darkstat : main</title>%s</head><body>%s<br>"
		      "<table border=\"0\" cellspacing=\"0\" cellpadding=\"1\">",
		css, linkbox);
	
	/* packets */
	x_packets = add_commas( bignum_from_64bit(num_packets), 1 );
	c_append(cnt, "<tr><td align=\"right\"><b>Packets captured:</b></td>"
		"<td>%s</td></tr>", x_packets);
	free(x_packets);

	/* data */
	x_data = add_commas( bignum_from_64bit(total_data), 1 );
	c_append(cnt, "<tr><td align=\"right\"><b>Data off the wire:</b></td>"
		"<td>%s bytes</td></tr>", x_data);
	free(x_data);

	/* uptime */
#define S_MIN 60
#define S_HR (S_MIN*60)
#define S_DAY (S_HR*24)

	days = uptime / S_DAY;
	uptime -= days * S_DAY;

	hours = uptime / S_HR;
	uptime -= hours * S_HR;

	minutes = uptime / S_MIN;
	uptime -= minutes * S_MIN;

#undef S_DAY
#undef S_HR
#undef S_MIN

	c_append(cnt, "<tr><td align=\"right\"><b>Counting:</b></td><td>");
	if (days)
		c_append(cnt, "%u days ", days);
	if (days || hours)
		c_append(cnt, "%u hours ", hours);
	if (days || hours || minutes)
		c_append(cnt, "%u minutes ", minutes);
	c_append(cnt, "%u seconds</td></tr>", uptime);

	/* since */
	c_append(cnt, "<tr><td align=\"right\"><b>Running since:</b></td><td>%s</td></tr>"
		      "</table><br>",
		asctime(localtime(&t_start)) );

	pthread_mutex_lock(&graph_mutex);
	html_graph(cnt, "Last 60 seconds", 400, 200,
			graph_sec_in, graph_sec_out, pos_sec, GRAPH_SECS);
	pthread_mutex_unlock(&graph_mutex);

	c_append(cnt, "</body></html>\n");

	return cnt;
}



content *www_graphs(SOCKET incoming)
{
	content *cnt = c_new();

	c_append(cnt,
		"<html><head><title>darkstat : graphs</title>%s</head><body>%s<br>"
		"<table border=\"0\">"
		"<tr><td align=\"center\">",
		css, linkbox);

	pthread_mutex_lock(&graph_mutex);
	html_graph(cnt, "Last 60 seconds", 300, 200,
			graph_sec_in, graph_sec_out, pos_sec, GRAPH_SECS);
	c_append(cnt, "</td><td align=\"center\">");
	html_graph(cnt, "Last 60 minutes", 300, 200,
			graph_min_in, graph_min_out, pos_min, GRAPH_MINS);
	c_append(cnt, "</td></tr><tr><td align=\"center\">");
	html_graph(cnt, "Last 24 hours", 300, 200,
			graph_hr_in, graph_hr_out, pos_hr, GRAPH_HRS);
	c_append(cnt, "</td><td align=\"center\">");
	html_graph(cnt, "Last 30 days", 300, 200,
			graph_day_in, graph_day_out, pos_day, GRAPH_DAYS);
	pthread_mutex_unlock(&graph_mutex);

	c_append(cnt,
		"</td></tr></table>"
		"</body></html>\n");

	return cnt;
}



content *www_hosts(SOCKET incoming, sort_type st, int limited)
{
	int i;
	content *cnt = c_new();
	host_record **list;
	int listend;

	c_append(cnt,
		"<html><head><title>darkstat : hosts</title>%s</head><body>%s<br>"
		"Hosts (sorted by %s",
		css, linkbox,
		(st==IN)?"incoming":(st==OUT)?"outgoing":(st==TOTAL)?"total":"ERROR");

	pthread_mutex_lock(&db_mutex);
	if (limited)
		c_append(cnt, ", top %d", min(HOST_REPORT_LIMIT, host_db.used));

	c_append(cnt, ")<br><br>"
		"<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\">"
		"<tr><td bgcolor=\"#000000\">"
		"<table border=\"0\" cellspacing=\"0\" cellpadding=\"2\"><tr>"
		"<td><font color=\"#FFFFFF\"><b>IP</b></font></td>"
		"<td><font color=\"#FFFFFF\"><b>Hostname</b></font></td>"
		"<td><font color=\"#FFFFFF\">"
			"<b><a class=\"black\" href=\"hosts-in.html\">In</a></b>"
			" (<a class=\"black\" href=\"hosts-in-full.html\">full</a>)</font></td>"
		"<td><font color=\"#FFFFFF\">"
			"<b><a class=\"black\" href=\"hosts-out.html\">Out</a></b>"
			" (<a class=\"black\" href=\"hosts-out-full.html\">full</a>)</font></td>"
		"<td><font color=\"#FFFFFF\">"
			"<b><a class=\"black\" href=\"hosts-total.html\">Total</a></b>"
			" (<a class=\"black\" href=\"hosts-total-full.html\">full</a>)</font></td>"
		"</tr>");

	list = host_db_sort(st);

	/* you want the full thing or not? */
	if (limited)
		listend = host_db.used-HOST_REPORT_LIMIT;
	else
		listend = 0;
	if (listend < 0) listend = 0;

	for (i=host_db.used-1; i>=listend; i--)
	{
		host_record *h;
		char *data_in, *data_out, *total;
		char *cellbg;
		i64 t64;
		
		if (!(i % 2))
			cellbg = "\"#F0F0F0\"";
		else
			cellbg = "\"#E0E0E0\"";

		/* grab the record */	
		h = list[i];

		data_in = add_commas( bignum_from_64bit(h->data_in), 1);
		data_out = add_commas( bignum_from_64bit(h->data_out), 1);

		t64 = h->data_in;
		add_64bit(&t64, h->data_out.lo);
		t64.hi += h->data_out.hi;
		total = add_commas( bignum_from_64bit(t64), 1);

		c_append(cnt,
			"<tr>"
			"<td bgcolor=%s>%d.%d.%d.%d</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td></tr>",
			cellbg, (h->ip_addr >> 24) & 255,
				(h->ip_addr >> 16) & 255,
				(h->ip_addr >> 8) & 255,
				 h->ip_addr & 255,
			cellbg, h->hostname?h->hostname:"&nbsp;",
			cellbg, data_in,
			cellbg, data_out,
			cellbg, total);

		free(data_in);
		free(data_out);
		free(total);
	}
	free(list);
	pthread_mutex_unlock(&db_mutex);

	c_append(cnt,
		"</table>"
		"</td></tr></table>"
		"</body></html>\n");

	return cnt;
}



content *www_ports(SOCKET incoming, sort_type st, int limited)
{
	content *cnt = c_new();
	int i;
	port_record **list;
	int listend;

	c_append(cnt,
		"<html><head><title>darkstat : ports</title>%s</head><body>%s<br>"
		"Ports (TCP, sorted by %s",
		css, linkbox, 
		(st==IN)?"incoming":(st==OUT)?"outgoing":(st==TOTAL)?"total":"ERROR");

	pthread_mutex_lock(&db_mutex);

	if (limited)
		c_append(cnt, ", top %d", min(PORT_REPORT_LIMIT, port_db.used));
		
	c_append(cnt,
		")<br><br>"
		"<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\">"
		"<tr><td bgcolor=\"#000000\">"
		"<table border=\"0\" cellspacing=\"1\" cellpadding=\"2\">"
		"<tr><td bgcolor=\"#000000\"><font color=\"#FFFFFF\"><b>Port</b></font></td>"
		"<td><font color=\"#FFFFFF\">"
			"<b><a class=\"black\" href=\"ports-in.html\">In</a></b>"
			" (<a class=\"black\" href=\"ports-in-full.html\">full</a>)</font></td>"
		"<td><font color=\"#FFFFFF\">"
			"<b><a class=\"black\" href=\"ports-out.html\">Out</a></b>"
			" (<a class=\"black\" href=\"ports-out-full.html\">full</a>)</font></td>"
		"<td><font color=\"#FFFFFF\">"
			"<b><a class=\"black\" href=\"ports-total.html\">Total</a></b>"
			" (<a class=\"black\" href=\"ports-total-full.html\">full</a>)</font></td>"
		"</tr>");

	list = port_db_sort(st);

	/* you want the full thing or not? */
	if (limited)
		listend = port_db.used-PORT_REPORT_LIMIT;
	else
		listend = 0;
	if (listend < 0) listend = 0;

	for (i=port_db.used-1; i>=listend; i--)
	{
		port_record *p;
		char *data_in, *data_out, *total;
		char *cellbg;

		if (!(i % 2))
			cellbg = "\"#F0F0F0\"";
		else
			cellbg = "\"#E0E0E0\"";

		/* grab the record */	
		p = list[i];

		data_in = bignum_from_64bit(p->data_in);
		data_out = bignum_from_64bit(p->data_out);
		total = add_commas( add_bignum(data_in, data_out), 1);
		data_in = add_commas(data_in, 1);
		data_out = add_commas(data_out, 1);

		c_append(cnt,
			"<tr><td bgcolor=%s>%d</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td></tr>",
			cellbg,p->port, cellbg,data_in, cellbg,data_out, cellbg,total);

		free(data_in);
		free(data_out);
		free(total);
	}
	pthread_mutex_unlock(&db_mutex);

	c_append(cnt,
		"</table>"
		"</td></tr></table>"
		"</body></html>\n");

	return cnt;
}



content *www_protocols(SOCKET incoming)
{
	int i;
	content *cnt = c_new();

	c_append(cnt,
		"<html><head><title>darkstat : protocols</title>%s</head><body>%s<br>"
		"<table border=\"0\" cellpadding=\"1\" cellspacing=\"0\">"
		"<tr><td bgcolor=\"#000000\">"
		"<table border=\"0\" cellspacing=\"1\" cellpadding=\"2\"><tr>"
		"<td colspan=\"2\" bgcolor=\"#000000\">"
			"<font color=\"#FFFFFF\"><b>Protocol</b></font></td>"
		"<td bgcolor=\"#000000\"><font color=\"#FFFFFF\"><b>In</b></font></td>"
		"<td bgcolor=\"#000000\"><font color=\"#FFFFFF\"><b>Out</b></font></td>"
		"<td bgcolor=\"#000000\"><font color=\"#FFFFFF\"><b>Total</b></font></td></tr>",
		css, linkbox);

	pthread_mutex_lock(&db_mutex);

	for (i=0; i<proto_db.used; i++)
	{
		proto_record *p;
		char *data_in, *data_out, *total;
		char *cellbg;

		if (!(i % 2))
			cellbg = "\"#F0F0F0\"";
		else
			cellbg = "\"#E0E0E0\"";

		/* grab the record */	
		p = (proto_record*)db_get(proto_db, i);

		data_in = bignum_from_64bit(p->data_in);
		data_out = bignum_from_64bit(p->data_out);
		total = add_commas( add_bignum(data_in, data_out), 1);

		data_in = add_commas(data_in, 1);
		data_out = add_commas(data_out, 1);

		c_append(cnt,
			"<tr><td bgcolor=%s align=\"right\">%d</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td>"
			"<td bgcolor=%s>%s</td></tr>\n",
			cellbg,p->protocol, cellbg,proto_name(p->protocol),
			cellbg,data_in, cellbg,data_out, cellbg,total);

		free(data_in);
		free(data_out);
		free(total);
	}
	pthread_mutex_unlock(&db_mutex);

	c_append(cnt,
		"</table>"
		"</td></tr></table>"
		"</body></html>\n");

	return cnt;
}



inline content *new_gif(void)
{
	content *cnt = c_new();

	assert(GIF_SIZE <= cnt->pool);
	memcpy(cnt->buf, gif, GIF_SIZE);
	return cnt;
}



content *www_unimplemented()
{
	content *c = c_new();
	c_append(c, "<html><body>"
		    "<h1>Not Implemented</h1>"
		    "Whatever the heck you just requested, I can't generate."
		    "</body></html>");

	return c;
}



content *www_notfound()
{
	content *c = c_new();
	c_append(c, "<html><body>"
		    "<h1>Not Found</h1>"
		    "Whatever you just requested, I don't have."
		    "</body></html>");

	return c;
}



#define SCOREBOARD 10
SOCKET		sb_socket[SCOREBOARD];
pthread_t	sb_thread[SCOREBOARD];
int		sb_complete[SCOREBOARD];

#define MAX_REQ_SIZE 8192

void handle_connection(void *arg)
{
	content *hdr, *cnt;
	char buf[4096], request[MAX_REQ_SIZE];
	int recvd, is_gif = 0;
	int select_ret, req_incomplete = 1;
	int slot = (int)arg;
	SOCKET incoming = sb_socket[slot];

	assert(incoming > 0);
	request[0] = 0;
	do
	{
		recvd = recv(incoming, buf, 4096, 0);
		buf[recvd] = 0;

		if (strlen(request) + strlen(buf) >= MAX_REQ_SIZE)
		{
			if (!quiet_mode)
				printf("WWW(%d): Request exceeded size of %d.\n",
						getpid(), MAX_REQ_SIZE);
			
			shutdown(incoming, 2);
			close(incoming);
			pthread_exit(0);
		}

		strcat(request, buf);

		if (
			(strcmp(request+strlen(request)-2, "\n\n") == 0) ||
			(strcmp(request+strlen(request)-2, "\r\r") == 0) ||
			(strcmp(request+strlen(request)-4, "\r\n\r\n") == 0) ||
			(strcmp(request+strlen(request)-4, "\n\r\n\r") == 0) )
				req_incomplete = 0;

	} while (recvd && req_incomplete);

	/* did the browser bail out? */
	if (req_incomplete)
	{
		if (!quiet_mode)
			printf("WWW(%d): Incomplete request sent.\n", getpid());
		close(incoming);
		pthread_exit(0);
	}

	shutdown(incoming, 0); /* no more recv */

	/* recv whatever's left */
	do {
		struct timeval timeout;
		fd_set fd;
		
		FD_ZERO(&fd);
		FD_SET(incoming, &fd);
		
		timeout.tv_sec = 2;
		timeout.tv_usec = 0;

		select_ret = select(incoming+1, &fd, NULL, NULL, &timeout);
	}
	while ( (select_ret > 0) && (recv(incoming, buf, 4096, 0) > 0) );

#define CMP(x) (memcmp(request,x,strlen(x)) == 0)

	hdr = c_new();

	if (!CMP("GET ")) {
		c_append(hdr, "HTTP/1.0 501 Not Implemented\r\n");
		cnt = www_unimplemented();
	} else
	if (CMP("GET / ") ||
		CMP("GET /index.html "))	cnt = www_frontpage(incoming); else

	if CMP("GET /hosts-total.html ")	cnt = www_hosts(incoming,TOTAL,1); else
	if CMP("GET /hosts-total-full.html ")	cnt = www_hosts(incoming,TOTAL,0); else
	if CMP("GET /hosts-in.html ")		cnt = www_hosts(incoming,IN,1); else
	if CMP("GET /hosts-in-full.html ")	cnt = www_hosts(incoming,IN,0); else
	if CMP("GET /hosts-out.html ")		cnt = www_hosts(incoming,OUT,1); else
	if CMP("GET /hosts-out-full.html ")	cnt = www_hosts(incoming,OUT,0); else

	if CMP("GET /ports-total.html ")	cnt = www_ports(incoming,TOTAL,1); else
	if CMP("GET /ports-total-full.html ")	cnt = www_ports(incoming,TOTAL,0); else
	if CMP("GET /ports-in.html ")		cnt = www_ports(incoming,IN,1); else
	if CMP("GET /ports-in-full.html ") 	cnt = www_ports(incoming,IN,0); else
	if CMP("GET /ports-out.html ")		cnt = www_ports(incoming,OUT,1); else
	if CMP("GET /ports-out-full.html ")	cnt = www_ports(incoming,OUT,0); else

	if CMP("GET /protocols.html ")		cnt = www_protocols(incoming); else
	if CMP("GET /graphs.html ")		cnt = www_graphs(incoming); else

	if CMP("GET /i/")
	{
		cnt = new_gif();
		is_gif = 1;

		if CMP("GET /i/r ")	cnt->buf[GIF_RED] = 0xFF; else
		if CMP("GET /i/b ")	cnt->buf[GIF_BLUE] = 0xFF; else
		if CMP("GET /i/m ")	cnt->buf[GIF_RED] = cnt->buf[GIF_BLUE] = 0xFF; else
		/*if CMP("GET /i/x ")*/	cnt->buf[GIF_PIXEL] = GIF_TRANSPARENT;
	} else {
		c_append(hdr, "HTTP/1.0 200 OK\r\n");
		cnt = www_notfound();
	}

#undef CMP

	if (!strlen(hdr->buf)) c_append(hdr, "HTTP/1.0 200 OK\r\n");
	if (!is_gif) c_append(hdr, "Cache-Control: no-cache\r\n");
	if (is_gif)
		c_append(hdr, "Content-Type: image/gif\r\n");
	else
		c_append(hdr, "Content-Type: text/html\r\n");
	c_append(hdr,	"Content-Length: %d\r\n"
			"Connection: close\r\n"
			"Server: " PACKAGE " v" VERSION "\r\n"
			"\r\n",
			(is_gif) ? GIF_SIZE : strlen(cnt->buf));

	/* send() raises SIGPIPE! */
	assert(hdr != NULL);
	assert(hdr->buf != NULL);
	assert(strlen(hdr->buf) > 0);
	send(incoming, hdr->buf, strlen(hdr->buf), MSG_NOSIGNAL);
	c_delete(&hdr);
	
	if (is_gif)
		send(incoming, cnt->buf, GIF_SIZE, MSG_NOSIGNAL);
	else
		send(incoming, cnt->buf, strlen(cnt->buf), MSG_NOSIGNAL);
	c_delete(&cnt);

	shutdown(incoming, 1); /* no more sends */
	sb_complete[slot] = 1;

	pthread_exit(0); /* not necessary? */
}



SOCKET sockin;

void www_main(void *ptr)
{
	struct sockaddr_in addrin;
	int sockopt = 1, i;

	/* reset scoreboard */
	for (i=0; i<SCOREBOARD; i++)
	{
		sb_socket[i] = -1;
		sb_complete[i] = 0;
	}

	/* create the incoming socket */
	sockin = socket(AF_INET, SOCK_STREAM, 0);
	if (sockin < 0) error("Problem creating socket");

	/* reuse address */
	if (setsockopt(sockin, SOL_SOCKET, SO_REUSEADDR, (char*)&sockopt, sizeof(sockopt)) < 0)
		error("WWW: Can't REUSEADDR");

	/* fill out a sockaddr struct */
	addrin.sin_family = AF_INET;
	addrin.sin_port = htons(webport);
	addrin.sin_addr.s_addr = INADDR_ANY;
	bzero(&(addrin.sin_zero), 8);

	/* bind sockin to the incoming port */
	if (bind(sockin, (struct sockaddr *)&addrin, sizeof(struct sockaddr)) < 0)
		error("Problem binding socket");

	/* listen on the socket */
	if (listen(sockin, MAXLISTEN) < 0)
		error("Problem listen()ing to socket");

	pid_www = getpid();
	printf("WWW: Thread is awake and waiting for connections. (PID:%d)\n", pid_www);

	while (pid_www)
	{
		int select_ret;
		struct timeval timeout;
		fd_set acceptfds;

		/* wait for an incoming connection on sockin */
		FD_ZERO(&acceptfds);
		FD_SET(sockin, &acceptfds);

		timeout.tv_sec = POLLDELAY;
		timeout.tv_usec = 0;

		select_ret = select(sockin+1, &acceptfds, NULL, NULL, &timeout);

		if (select_ret > 0) {
			struct sockaddr_in addrin;
			int sin_size;
			SOCKET incoming;
			int slot = -1;
	
			/* do we have a free slot? */
			for (i=0; (i<SCOREBOARD) && (slot < 0); i++)
				if (sb_socket[i] == -1) slot = i;

			/* we do! */
			if (slot >= 0) {
				sin_size = sizeof(struct sockaddr);
				incoming = accept(sockin, (struct sockaddr *)&addrin, &sin_size);

				/* changed your mind? <-- happens often when the
				 * web interface is being hammered with requests
				 */
				if (incoming < 0)
				{
					if (!quiet_mode)
						printf("WWW: Lost a connection in accept()\n");
				}
				else
				{
					/* details about incoming connection */
					if (!quiet_mode)
						printf("WWW: Got a connection from %s:%d\n",
							inet_ntoa(addrin.sin_addr),
							ntohs(addrin.sin_port));

					sb_socket[slot] = incoming;
					sb_complete[slot] = 0;
					if (pthread_create(&sb_thread[slot], NULL,
						(void*)&handle_connection, (void*)slot) == -1)
							error("Thread creation error.");
				}
			} /* else freakout because there's no slots free? */
		}
	
		/* see if any children are finished */
		for (i=0; i<SCOREBOARD; i++)
		if (sb_complete[i])
		{
			pthread_join(sb_thread[i], NULL); /* make sure thread is dead */
			sb_complete[i] = 0;	/* reset completeness */
			close(sb_socket[i]);	/* we opened it, we close it. */
			sb_socket[i] = -1;	/* reset socket */
		}
			
		fflush(stdout);
	}

	pthread_exit(0);
}



void www_shutdown(void)
{
	close(sockin);
}

